{-# OPTIONS -cpp -#include windows.h -#include commctrl.h #-}

-- ----------------------------------------------------------------------------
-- |
-- Module    : ProjectModuleBrowser
-- Author    : Krasimir Angelov <kr.angelov@gmail.com>
-- Copyright : (c) Krasimir Angelov, All Rights Reserved
-- 
-- ----------------------------------------------------------------------------

module ProjectModuleBrowser
	( newProjectModuleBrowser
	) where

import Wtypes
import VsTypes
import SiteServices
import FilePath
import ProjectState
import ProjectContent
import TemplateGenerator
import ServProv
import VsProject( IVsProjectFactory(..), iidIVsProjectFactory, iidIVsUIHierarchy, iidIVsHierarchy, IVsHierarchy(..), IVsWindowFrame )
import VsProjectProxy( mkIVsProjectFactory_vtbl )
import VsClassView ( IVsLibraryMgr,   iidIVsLibraryMgr 
                   , IVsLibrary,      iidIVsLibrary
                   , IVsLiteTreeList, iidIVsLiteTreeList
                   , IVsObjectList,   iidIVsObjectList
                   , LIB_CHECKSTATE(..)
                   , LIB_LISTTYPE(..)
                   , LIB_PERSISTTYPE(..)
                   , LIB_LISTFLAGS(..)
                   , LIB_FLAGS(..)                   
                   , LIB_CATEGORY(..)
                   , LIBCAT_MODIFIERTYPE(..)
                   , LIBCAT_MEMBERTYPE(..)
                   , LIBCAT_MEMBERACCESS(..)
                   , LIBCAT_CLASSTYPE(..)
                   , LIBCAT_CLASSACCESS(..)
                   , LIBCAT_ACTIVEPROJECT(..)
                   , LIBCAT_VISIBILITY(..)
                   , VSOBSEARCHCRITERIA(..)
                   , VSBROWSECONTAINER(..)
                   , VSCOMPONENTSELECTORDATA
                   , VSTREELISTITEMCHANGE(..)
                   , VSTREEITEMCHANGESMASK(..)
                   , VSTREETEXTOPTIONS(..)
                   , VSTREEFLAGS(..)
                   , VSTREEDISPLAYDATA(..)
                   , VSTREEDISPLAYMASK(..)
                   , VSTREEDISPLAYSTATE(..)
                   , VSOBJGOTOSRCTYPE(..)
                   )
import VsClassViewProxy
import BabelServiceLib(iconGroupSize, ScopeIconGroup(..), ScopeIconItem(..))

-- std libraries
import Control.Exception(handle)
import Control.Monad(unless)
import Data.Array
import Data.Char(isSpace)
import Data.Bits((.&.))
import Data.Word
import Data.IORef
import Foreign.Ptr(Ptr(..))
import System.IO.Unsafe(unsafePerformIO)

-- Cabal
import Distribution.PackageDescription

-- COM
import Com
import ComServ
import WideString
import ComException(e_FAIL, e_UNEXPECTED, e_NOTIMPL)

import VsGHC
import SrcLoc(noSrcLoc, srcLocLine, srcLocCol)
import Outputable
import PprTyThing
import Name(getName)
import Debug.Trace
#include "../../vs_haskell_ui/resource.h"

data BrowserState = BrowserState
        { browserImgList   :: {-# UNPACK #-} !HIMAGELIST
        , browserItemId    :: VSITEMID
        , browserProject   :: ProjectState
        , browserTyThings  :: Array Int TyThing
        , browserInstances :: Array Int Instance
        , browserIsExportedName :: Name -> Bool
        , browserSortWSs   :: Array Int WideString
        , browserWString   :: {-# UNPACK #-} !(IORef WideString)
        , browserInstWS    :: WideString
        }

newProjectModuleBrowser :: HIMAGELIST -> VSITEMID -> ProjectState -> IO (IVsObjectList ())
newProjectModuleBrowser imageList itemid projState = do
  (tyThings,instances,isExported) <- getItemTyThings itemid (prjContent projState)
  lstWStrings <- mapM getSortName (elems tyThings)
  let sortWSs = listArray (0,length lstWStrings-1) lstWStrings
  refWString <- newIORef nullWideString
  instWS <- stringToWide "\255"
  let st = BrowserState 
	          { browserImgList   = imageList
	          , browserItemId    = itemid
	          , browserProject   = projState
	          , browserTyThings  = tyThings
	          , browserInstances = instances
	          , browserIsExportedName = isExported
	          , browserSortWSs   = sortWSs
	          , browserWString   = refWString
	          , browserInstWS    = instWS
	          }
  createComInstance "" st (releaseModuleBrowser st) ifaces_ProjectModuleBrowser iidIVsObjectList
  where
    getSortName :: TyThing -> IO WideString
    getSortName tyThing =
      stringToWide (stripSpaces (showSDocUnqual (ppr (getName tyThing))))

stripSpaces []     = []
stripSpaces (c:cs)
    | isSpace c = ' ' : dropWhile isSpace (stripSpaces cs)
    | otherwise = c   : stripSpaces cs

releaseModuleBrowser :: BrowserState -> IO ()
releaseModuleBrowser st = do
  woldname <- readIORef (browserWString st)
  unless (woldname == nullWideString) $
    freeWString woldname
  mapM_ freeWString (elems (browserSortWSs st))
  freeWString (browserInstWS st)

getFlags :: BrowserState -> IO VSTREEFLAGS
getFlags st = return (VSTREEFLAGSList__ [TF_NOEXPANSION,TF_NORELOCATE,TF_NOINSERTDELETE,TF_NOREALIGN,TF_NOEFFECTS,TF_NOCOLORS])
 
getItemCount :: BrowserState -> IO ULONG
getItemCount st = do
  return (fromIntegral (rangeSize (bounds (browserTyThings  st)) + 
                        rangeSize (bounds (browserInstances st))))

getExpandedList _ _ = coFailWithHR e_NOTIMPL "getExpandedList isn't implemented"
locateExpandedList _ _ = coFailWithHR e_NOTIMPL "locateExpandedList isn't implemented"
onClose _ = coFailWithHR e_NOTIMPL "onClose isn't implemented"

getText :: ULONG -> VSTREETEXTOPTIONS -> BrowserState -> IO WideString
getText index1 TTO_SORTTEXT st
  | inRange (bounds tyThings) index =
      return (browserSortWSs st ! index)
  | inRange (bounds instances) index =
      return (browserInstWS st)
  | otherwise = return nullWideString
  where
    index = fromIntegral index1
    tyThings  = browserTyThings  st
    instances = browserInstances st

getText index1 tto st = do
  woldname <- readIORef (browserWString st)
  unless (woldname == nullWideString) $
    freeWString woldname
  let index = fromIntegral index1
      tyThings  = browserTyThings  st
      instances = browserInstances st
      doc | inRange (bounds tyThings) index =
              pprTyThingHdr True (tyThings ! index)
          | inRange (bounds instances) index =
              pprInstanceHdr (instances ! index)
          | otherwise = empty
  wname <- stringToWide (stripSpaces (showSDocUnqual doc))
  writeIORef (browserWString st) wname
  return wname

getTipText _ _ _ = coFailWithHR e_NOTIMPL "getTipText isn't implemented"

getExpandable :: ULONG -> BrowserState -> IO BOOL
getExpandable index st = return 0

getDisplayData :: ULONG -> BrowserState -> IO VSTREEDISPLAYDATA
getDisplayData index' st = return $!
  (VSTREEDISPLAYDATA { mask = VSTREEDISPLAYMASKList__ [TDM_IMAGE,TDM_SELECTEDIMAGE]
                     , state = VSTREEDISPLAYSTATEList__ []
                     , stateMask = VSTREEDISPLAYSTATEList__ []
                     , image = imageIndex
                     , selectedImage = imageIndex
                     , hImageList = browserImgList st
                     , forceSelectStart = 0
                     , forceSelectLength = 0
                     })
  where
    imageIndex
      | inRange (bounds tyThings) index =
          case tyThing of
            ATyCon _             -> icon IconGroupType   iconMod
            ADataCon _           -> icon IconGroupType2  iconMod
            AnId id
              | isDictonaryId id -> icon IconGroupType3  iconMod
              | otherwise        -> icon IconGroupMethod iconMod
            AClass _             -> icon IconGroupClass  iconMod
      | inRange (bounds instances) index =
          icon IconGroupType3 IconItemPublic
      | otherwise = 0
      where
        index    = fromIntegral index'
        tyThings = browserTyThings st
        tyThing  = tyThings ! index
        instances= browserInstances st
        name     = Name.getName tyThing
        iconMod | browserIsExportedName st name = IconItemPublic
                | otherwise                     = IconItemInternal

    icon :: ScopeIconGroup -> ScopeIconItem -> USHORT
    icon group status = fromIntegral (fromEnum group) * fromIntegral (iconGroupSize)
                      + fromIntegral (fromEnum status)

updateCounter :: BrowserState -> IO (ULONG, VSTREEITEMCHANGESMASK)
updateCounter st = do
  let itemid    = browserItemId  st
      projState = browserProject st
  (counter, change) <- getItemUpdateCounter itemid (prjContent projState)
  return (counter,grfChange change)

getListChanges :: ULONG -> BrowserState -> IO [VSTREELISTITEMCHANGE]
getListChanges count st
  | count /= 1 = coFailWithHR e_UNEXPECTED "Invalid count"
  | otherwise  = do
      let itemid    = browserItemId  st
          projState = browserProject st
      (counter, change) <- getItemUpdateCounter itemid (prjContent projState)
      return [change]

toggleState index st = coFailWithHR e_NOTIMPL "getCapabilities isn't implemented"
getCapabilities _ = coFailWithHR e_NOTIMPL "getCapabilities isn't implemented"

getList :: ULONG -> LIB_LISTTYPE -> LIB_LISTFLAGS -> Maybe VSOBSEARCHCRITERIA -> BrowserState -> IO (IVsObjectList ())
getList _ _ _ _ _ = coFailWithHR e_UNEXPECTED "Unsupported list type"

getCategoryField :: ULONG -> LIB_CATEGORY -> BrowserState -> IO DWORD
getCategoryField index' LC_MEMBERTYPE st
  | inRange (bounds tyThings)  index = return $!
      case tyThings ! index of
        ATyCon _   -> fromIntegral (fromEnum LCMT_TYPEDEF)
        ADataCon _ -> fromIntegral (fromEnum LCMT_ENUMITEM)
        AnId id    -> fromIntegral (fromEnum LCMT_FUNCTION)
        AClass _   -> fromIntegral (fromEnum LCMT_PROPERTY)
  | inRange (bounds instances) index = return $!
      fromIntegral (fromEnum LCMT_PROPERTY)
  where
    index     = fromIntegral index'
    tyThings  = browserTyThings st
    instances = browserInstances st
getCategoryField index' LC_MEMBERACCESS st
  | inRange (bounds tyThings) index &&
    not (isExported name)           =  return $! fromIntegral (fromEnum LCMA_PACKAGE)
  | otherwise                       =  return $! fromIntegral (fromEnum LCMA_PUBLIC)
  where
    index      = fromIntegral index'
    tyThings   = browserTyThings st
    name       = Name.getName (tyThings ! index)
    isExported = browserIsExportedName st
getCategoryField index _ st = return 0


getExpandable2 _ _ _ = coFailWithHR e_NOTIMPL "getExpandable2 isn't implemented"
getNavigationInfo _ _ _ = coFailWithHR e_NOTIMPL "getNavigationInfo isn't implemented"
locateNavigationInfo _ _ _ _ = coFailWithHR e_NOTIMPL "locateNavigationInfo isn't implemented"
getBrowseObject _ _ = coFailWithHR e_NOTIMPL "getBrowseObject isn't implemented"
getUserContext _ _ = coFailWithHR e_NOTIMPL "getUserContext isn't implemented"
showHelp _ _ = coFailWithHR e_NOTIMPL "showHelp isn't implemented"
getSourceContext _ _ = coFailWithHR e_NOTIMPL "getSourceContext isn't implemented"

countSourceItems :: ULONG -> BrowserState -> IO (IVsHierarchy (), VSITEMID, ULONG)
countSourceItems index st = do
  iHier <- queryInterface iidIVsHierarchy (prjSelf (browserProject st))
  return (iHier, browserItemId st, 1)

getMultipleSourceItems _ _ _ _ = coFailWithHR e_NOTIMPL "getMultipleSourceItems isn't implemented"

canGoToSource :: ULONG -> VSOBJGOTOSRCTYPE -> BrowserState -> IO BOOL
canGoToSource index GS_DEFINITION st = return 1
canGoToSource index _             st = return 0

goToSource :: ULONG -> VSOBJGOTOSRCTYPE -> BrowserState -> IO ()
goToSource index' srcType st = do
  let itemid    = browserItemId  st
      projState = browserProject st
  fpath <- getItemFilePath itemid (prjLocation projState) (prjContent projState)
  iFrame <- withQueryInterface iidIVsUIHierarchy (prjSelf projState) $ \uiHier ->
              openFile fpath uiHier itemid (prjServices projState)
  let srcLoc | inRange (bounds tyThings)  index = 
                  nameSrcLoc (Name.getName (tyThings ! index))
             | inRange (bounds instances) index =
                  nameSrcLoc (Name.getName (instances ! index))
             | otherwise = noSrcLoc
  iFrame <- marshallIUnknown iFrame
  windowFrameSetCarret iFrame (fromIntegral (srcLocLine srcLoc - 1)) (fromIntegral (srcLocCol srcLoc))
  return ()
  where
    index      = fromIntegral index'
    tyThings   = browserTyThings  st
    instances  = browserInstances st

foreign import ccall "WindowFrameSetCarret" windowFrameSetCarret :: Ptr (IVsWindowFrame ()) -> LONG -> LONG -> IO HRESULT

getContextMenu _ _ = coFailWithHR e_NOTIMPL "getContextMenu isn't implemented"
queryDragDrop _ _ _ _ _ = coFailWithHR e_NOTIMPL "queryDragDrop isn't implemented"
doDragDrop _ _ _ _ _ = coFailWithHR e_NOTIMPL "doDragDrop isn't implemented"
canRename _ _ _ = coFailWithHR e_NOTIMPL "canRename isn't implemented"
doRename _ _ _ _ = coFailWithHR e_NOTIMPL "doRename isn't implemented"
canDelete _ _ = coFailWithHR e_NOTIMPL "canDelete isn't implemented"
doDelete _ _ _ = coFailWithHR e_NOTIMPL "doDelete isn't implemented"
fillDescription _ _ _ _ = coFailWithHR e_NOTIMPL "fillDescription isn't implemented"
enumClipboardFormats _ _ _ _ = coFailWithHR e_NOTIMPL "enumClipboardFormats isn't implemented"
getClipboardFormat _ _ _ _ _ = coFailWithHR e_NOTIMPL "getClipboardFormat isn't implemented"
getExtendedClipboardVariant _ _ _ _ = coFailWithHR e_NOTIMPL "getExtendedClipboardVariant isn't implemented"

iVsObjectList_vtbl :: ComServ.ComVTable (IVsObjectList ()) BrowserState
iVsObjectList_vtbl =
  unsafePerformIO (mkIVsObjectList_vtbl getFlags getItemCount getExpandedList locateExpandedList onClose getText getTipText getExpandable getDisplayData updateCounter getListChanges toggleState getCapabilities getList getCategoryField getExpandable2 getNavigationInfo locateNavigationInfo getBrowseObject getUserContext showHelp getSourceContext countSourceItems getMultipleSourceItems canGoToSource goToSource getContextMenu queryDragDrop doDragDrop canRename doRename canDelete doDelete fillDescription enumClipboardFormats getClipboardFormat getExtendedClipboardVariant)

ifaces_ProjectModuleBrowser :: [ComInterface BrowserState]
ifaces_ProjectModuleBrowser =
  [ ComServ.mkIface iidIVsLiteTreeList
                    iVsObjectList_vtbl
  , ComServ.mkIface iidIVsObjectList
                    iVsObjectList_vtbl
  ]

foreign import stdcall unsafe "ImageList_LoadBitmap" loadImageList :: HINSTANCE -> Int -> Int -> Int -> Word32 -> IO HIMAGELIST
foreign import stdcall unsafe "ImageList_Destroy" destroyImageList :: HIMAGELIST -> IO BOOL
